/* Written by Yu-Fang Chen, Richard Mayr, and Chih-Duo Hong               */
/* Copyright (c) 2010                  	                                  */
/*                                                                        */
/* This program is free software; you can redistribute it and/or modify   */
/* it under the terms of the GNU General Public License as published by   */
/* the Free Software Foundation; either version 2 of the License, or      */
/* (at your option) any later version.                                    */
/*                                                                        */
/* This program is distributed in the hope that it will be useful,        */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */
/*                                                                        */
/* You should have received a copy of the GNU General Public License      */
/* along with this program; if not, write to the Free Software            */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA*/

package algorithms;

import java.lang.management.ManagementFactory;
import java.lang.management.ThreadMXBean;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import automata.FAState;
import automata.FiniteAutomaton;
import comparator.StatePairComparator;
import comparator.SuperGraphComparator;
import datastructure.Arc;
import datastructure.HashSet;
import datastructure.OneToOneTreeMap;
import datastructure.Pair;



/**
 * 
 * @author Yu-Fang Chen
 * 
 */
public class InclusionOpt2 extends Thread{
	public boolean backward=true;
	public boolean C1=true;
	public boolean C2=true;
	public boolean opt2=true;
	public boolean quotient=true;
	public boolean debug=true;
	public int removedCnt;
	public boolean included=true;

	
	TreeSet<Pair<Arc,TreeSet<Arc>>> R=new TreeSet<Pair<Arc,TreeSet<Arc>>>(new SuperGraphComparator());
	TreeSet<Pair<Arc,TreeSet<Arc>>> L=new TreeSet<Pair<Arc,TreeSet<Arc>>>(new SuperGraphComparator());
	
	private TreeMap<Integer,TreeSet<FAState>> Post=new TreeMap<Integer,TreeSet<FAState>>();
	private TreeMap<Integer,TreeSet<FAState>> NPost=new TreeMap<Integer,TreeSet<FAState>>();

	private TreeMap<String,TreeSet<Integer>> Tail=new TreeMap<String,TreeSet<Integer>>();
	private TreeMap<String,TreeSet<Integer>> Head=new TreeMap<String,TreeSet<Integer>>();
	private long runTime;
	private boolean stop=false;
	private Set<Pair<FAState,FAState>> frel, brel;
	FiniteAutomaton spec, system;
	String showSim(Set<Pair<FAState,FAState>> sim){
		String result="";
		Iterator<Pair<FAState,FAState>> sim_it=sim.iterator();
		while(sim_it.hasNext()){
			Pair<FAState,FAState> p=sim_it.next();
			if(p.getLeft().compareTo(p.getRight())!=0)
				result+=("("+p.getLeft()+","+p.getRight()+")");
		}
		return result;
	}
	void debug(String out){
		if(debug)
		System.out.println(out);
	}
	
	
	public InclusionOpt2(FiniteAutomaton system, FiniteAutomaton spec){
		this.spec=spec;
		this.system=system;
	}
	/**
	 * Simplify a finite automaton by merging simulation equivalent states
	 * @param fa: a finite automaton
	 * @param Sim: some simulation rel_specation over states in the spec automaton
	 * 
	 * @return an equivalent finite automaton
	 */
	private FiniteAutomaton quotient(FiniteAutomaton fa, Set<Pair<FAState,FAState>> rel) {
		FiniteAutomaton result=new FiniteAutomaton();
		result.name=fa.name;
		TreeMap<FAState,FAState> map=new TreeMap<FAState,FAState>();
		TreeMap<FAState,FAState> reducedMap=new TreeMap<FAState,FAState>();
		
		Iterator<FAState> state_it=fa.states.iterator();
		while(state_it.hasNext()){
			FAState state=state_it.next();
			map.put(state, state);
			Iterator<FAState> state_it2=fa.states.iterator();
			while(state_it2.hasNext()){
				FAState state2=state_it2.next();
				if(rel.contains(new Pair<FAState,FAState>(state,state2)) &&
					rel.contains(new Pair<FAState,FAState>(state2,state))){
					map.put(state,state2);
				}
			}			
		}

		FAState init=result.createState();
		reducedMap.put(map.get(fa.getInitialState()), init);
		result.setInitialState(init);

		state_it=fa.states.iterator();
		while(state_it.hasNext()){
			FAState state=state_it.next();
			if(!reducedMap.containsKey(map.get(state))){
				reducedMap.put(map.get(state), result.createState());
			}
			if(fa.F.contains(state)){
				result.F.add(reducedMap.get(map.get(state)));
			}
			Iterator<String> sym_it=state.nextIt();
			while(sym_it.hasNext()){
				String sym=sym_it.next();
				Iterator<FAState> to_it=state.getNext(sym).iterator();
				while(to_it.hasNext()){
					FAState to=to_it.next();
					if(!reducedMap.containsKey(map.get(to))){
						reducedMap.put(map.get(to), result.createState());
					}
					result.addTransition(reducedMap.get(map.get(state)), reducedMap.get(map.get(to)), sym);
				}
			}
		}
		Set<Pair<FAState,FAState>> newrel=new TreeSet<Pair<FAState,FAState>>(new StatePairComparator());
		Iterator<Pair<FAState,FAState>> sim_it=rel.iterator();
		while(sim_it.hasNext()){
			Pair<FAState,FAState> sim=sim_it.next();
			FAState left,right;
			if(sim.getLeft().getowner()==fa){
				left=reducedMap.get(map.get(sim.getLeft()));
			}else{
				left=sim.getLeft();
			}
			
			if(sim.getRight().getowner()==fa){
				right=reducedMap.get(map.get(sim.getRight()));
			}else{
				right=sim.getRight();
			}
			newrel.add(new Pair<FAState,FAState>(left,right));
		}
		rel.clear();
		rel.addAll(newrel);
		
		return result;
	}	

	private boolean C1(Pair<Arc,TreeSet<Arc>> g){
		if(!C1){
			return false;
		}
		Arc left=g.getLeft();
		if(left.getFrom()!=system.getInitialState().id){
			debug("C1("+g+")="+true);
			return true;
		}
		Iterator<Arc> right_it=g.getRight().iterator();
		while(right_it.hasNext()){
			Arc right=right_it.next();
			if(right.getFrom()==spec.getInitialState().id){
				if(frel.contains(new Pair<FAState,FAState>(new FAState(left.getTo(),system),new FAState(right.getTo(),spec)))){
					debug("C1("+g+")="+true);
					return true;
				}
			}
		}
		debug("C1("+g+")="+false);
		return false;
	}

	private TreeSet<FAState> getPost(int qid){
		TreeSet<FAState> post=Post.get(qid);
		if(post==null){
			post=new TreeSet<FAState>();
			FAState q=null;
			Iterator<FAState> st_it=system.states.iterator();
			while(st_it.hasNext()){
				FAState st=st_it.next();
				if(st.getID()==qid){
					q=st;
					break;
				}
			}
			Set<FAState> next=q.getNext();
			post.addAll(next);
			post.add(q);
			while(!next.isEmpty()){
				FAState s=next.iterator().next();
				next.remove(s);
				Iterator<FAState> toAdd_it=s.getNext().iterator();
				while(toAdd_it.hasNext()){
					FAState toAdd=toAdd_it.next();
					if(!post.contains(toAdd)){
						post.add(toAdd);
						next.add(toAdd);
					}
				}
			}
			Post.put(qid, post);
		}
		return post;
	}
	private boolean C2(Pair<Arc,TreeSet<Arc>> g){
		if(!C2){
			return false;
		}
		if(g.getLeft().getLabel()){	
			TreeSet<FAState> post=getPost(g.getLeft().getTo());
			Iterator<FAState> postst_it=post.iterator();
			while(postst_it.hasNext()){
				FAState postst=postst_it.next();
				if(frel.contains(new Pair<FAState,FAState>(new FAState(g.getLeft().getFrom(),system),postst))){
					debug("C2("+g+")="+false);
					return false;
				}
			}
		}else{
			TreeSet<FAState> post=NPost.get(g.getLeft().getTo());
			if(post==null){
				TreeSet<FAState> fpost=new TreeSet<FAState>();
				fpost.addAll(getPost(g.getLeft().getTo()));
				fpost.retainAll(system.F);
				post=new TreeSet<FAState>();
				
				Iterator<FAState> st_it=fpost.iterator();
				while(st_it.hasNext()){
					FAState st=st_it.next();
					post.addAll(getPost(st.id));
				}
				NPost.put(g.getLeft().getTo(), post);
			}
			Iterator<FAState> postst_it=post.iterator();
			while(postst_it.hasNext()){
				FAState postst=postst_it.next();
				if(frel.contains(new Pair<FAState,FAState>(postst,new FAState(g.getLeft().getFrom(),system)))){
					debug("C2("+g+")="+false);
					return false;
				}
			}
		}
		debug("C2("+g+")="+true);
		return true;
	}
	
	private boolean double_graph_test(Pair<Arc,TreeSet<Arc>> g, Pair<Arc,TreeSet<Arc>> h, int init){
		Arc g_arc=g.getLeft();
		Arc h_arc=h.getLeft();
		
		if(!(frel.contains(new Pair<FAState,FAState>(new FAState(h_arc.getFrom(),system),new FAState(h_arc.getTo(),system))) &&
			 frel.contains(new Pair<FAState,FAState>(new FAState(h_arc.getFrom(),system),new FAState(g_arc.getTo(),system))) &&
			 system.getInitialState().id==g_arc.getFrom() && h_arc.getLabel())
		){
			debug("RDGT("+g+"["+stringmap.get(g)+"],"+h+"["+stringmap.get(h)+"])="+true);

			return true;
		}else{
			boolean result;
			if(!backward)
				result = lasso_finding_test(g.getRight(),h.getRight(),init);
			else
				result = lasso_finding_test_plus(g.getRight(),h.getRight(),init);

			if(!result){
				debug("g_arc: "+g_arc);
				debug("h_arc: "+h_arc);
				debug("init: "+system.getInitialState());
				debug("final: "+system.F);
			}
			debug("RDGT("+g+"["+stringmap.get(g)+"],"+h+"["+stringmap.get(h)+"])="+result);

			return result;
		}
	}	
		
	
	private boolean lasso_finding_test(TreeSet<Arc> g, TreeSet<Arc> h, int init){
		if(!Head.containsKey(g.toString())){
			TreeSet<Integer> H=new TreeSet<Integer>();
			Iterator<Arc> arc_g_it=g.iterator();
			while(arc_g_it.hasNext()){
				Arc arc_g=arc_g_it.next();
				if(arc_g.getFrom()==init){
					H.add(arc_g.getTo());
				}
			}
			Head.put(g.toString(), H);
		}
		if(!Tail.containsKey(h.toString())){
			FiniteAutomaton fa=new FiniteAutomaton();
			OneToOneTreeMap<Integer,automata.FAState> st=new OneToOneTreeMap<Integer,automata.FAState>();
			Iterator<Arc> arc_h_it=h.iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				if(!st.containsKey(arc_h.getFrom()))
					st.put(arc_h.getFrom(), fa.createState());
				if(!st.containsKey(arc_h.getTo()))
					st.put(arc_h.getTo(), fa.createState());
				fa.addTransition(st.getValue(arc_h.getFrom()), st.getValue(arc_h.getTo()), arc_h.getLabel()?"1":"0");
			}
			SCC s=new SCC(fa);
			TreeSet<Integer> T=new TreeSet<Integer>();
			Iterator<FAState> s_it=s.getResult().iterator();
			while(s_it.hasNext()){
				T.add(st.getKey(s_it.next()));
			}
			int TailSize=0;
			TreeSet<Arc> isolatedArcs=h;
			while(TailSize!=T.size()){
				TailSize=T.size();
				TreeSet<Arc> isolatedArcsTemp=new TreeSet<Arc>();
				Iterator<Arc> arc_it=isolatedArcs.iterator();
				while(arc_it.hasNext()){
					Arc arc=arc_it.next();
					if(!T.contains(arc.getTo())){
						isolatedArcsTemp.add(arc);
					}else{
						T.add(arc.getFrom());
					}
				}
				isolatedArcs=isolatedArcsTemp;
			}
			Tail.put(h.toString(), T);
		}
		TreeSet<Integer> intersection = new TreeSet<Integer>();
		intersection.addAll(Head.get(g.toString()));
		intersection.retainAll(Tail.get(h.toString()));
		
		if(debug){
			if(intersection.isEmpty()){
				debug("g_graph:"+g+", Head: "+Head.get(g.toString()));
				debug("h_graph:"+h+", Tail: "+Tail.get(h.toString()));
			}
		}
		
		return !intersection.isEmpty();
	}

	private boolean lasso_finding_test_plus(TreeSet<Arc> g, TreeSet<Arc> h, int init){
		if(!Head.containsKey(g.toString())){
			TreeSet<Integer> H=new TreeSet<Integer>();
			Iterator<Arc> arc_g_it=g.iterator();
			while(arc_g_it.hasNext()){
				Arc arc_g=arc_g_it.next();
				if(arc_g.getFrom()==init){
					H.add(arc_g.getTo());
					Iterator<Pair<FAState,FAState>>brel_it=brel.iterator();
					while(brel_it.hasNext()){
						Pair<FAState,FAState> p=brel_it.next();
						if(p.getLeft().getowner()==spec&&p.getRight().getowner()==spec){
							if(arc_g.getTo()==p.getLeft().id){
								H.add(p.getRight().id);
							}
						}
					}

				}
			}
			Head.put(g.toString(), H);
		}
		if(!Tail.containsKey(h.toString())){
			FiniteAutomaton fa=new FiniteAutomaton();
			OneToOneTreeMap<Integer,automata.FAState> st=new OneToOneTreeMap<Integer,automata.FAState>();
			Iterator<Arc> arc_h_it=h.iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				if(!st.containsKey(arc_h.getFrom())){
					st.put(arc_h.getFrom(), fa.createState());
				}
				if(!st.containsKey(arc_h.getTo())){
					st.put(arc_h.getTo(), fa.createState());
				}
				fa.addTransition(st.getValue(arc_h.getFrom()), st.getValue(arc_h.getTo()), arc_h.getLabel()?"1":"0");

				Iterator<Pair<FAState,FAState>>brel_it=brel.iterator();
				while(brel_it.hasNext()){
					Pair<FAState,FAState> p=brel_it.next();
					if(p.getLeft().getowner()==spec&&p.getRight().getowner()==spec){
						if(!st.containsKey(p.getLeft().getID()))
							st.put(p.getLeft().getID(), fa.createState());
						if(!st.containsKey(p.getRight().getID()))
							st.put(p.getRight().getID(), fa.createState());
						fa.addTransition(st.getValue(p.getLeft().getID()), st.getValue(p.getRight().getID()), "0");
					}
				}
			}
			SCC s=new SCC(fa);
			TreeSet<FAState> Pre=new TreeSet<FAState>();
			TreeSet<FAState> Next=s.getResult();
			Pre.addAll(Next);
			while(!Next.isEmpty()){
				FAState state=Next.iterator().next();
				Next.remove(state);
				TreeSet<FAState> toAdd=new TreeSet<FAState>();
				toAdd.addAll(state.getPre());
				toAdd.removeAll(Pre);
				Pre.addAll(toAdd);
				Next.addAll(toAdd);
			}
			Next=null;
			Iterator<FAState> Pre_it=Pre.iterator();
			TreeSet<Integer> T=new TreeSet<Integer>();
			while(Pre_it.hasNext()){
				FAState state=Pre_it.next();
				T.add(st.getKey(state)%spec.states.size());
			}
			Tail.put(h.toString(), T);
		}
		TreeSet<Integer> intersection = new TreeSet<Integer>();
		intersection.addAll(Head.get(g.toString()));
		intersection.retainAll(Tail.get(h.toString()));
		
		if(debug){
			if(intersection.isEmpty()){
				debug("g_graph:"+g+", Head: "+Head.get(g.toString()));
				debug("h_graph:"+h+", Tail: "+Tail.get(h.toString()));
			}
		}
		
		return !intersection.isEmpty();
	}
/*	private boolean lasso_finding_test_plus_old(TreeSet<Arc> g, TreeSet<Arc> h, int init){
		if(!Head.containsKey(g.toString())){
			TreeSet<Integer> H=new TreeSet<Integer>();
			Iterator<Arc> arc_g_it=g.iterator();
			while(arc_g_it.hasNext()){
				Arc arc_g=arc_g_it.next();
				if(arc_g.getFrom()==init){
					H.add(arc_g.getTo());
				}
			}
			Head.put(g.toString(), H);
		}
		if(!Tail.containsKey(h.toString())){
			FiniteAutomaton fa=new FiniteAutomaton();
			OneToOneTreeMap<Integer,cav2010.automata.FAState> st=new OneToOneTreeMap<Integer,cav2010.automata.FAState>();
			Iterator<Arc> arc_h_it=h.iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				if(!st.containsKey(arc_h.getFrom())){
					st.put(arc_h.getFrom(), fa.createState());
					st.put(arc_h.getFrom()+spec.states.size(), fa.createState());
				}
				if(!st.containsKey(arc_h.getTo())){
					st.put(arc_h.getTo(), fa.createState());
					st.put(arc_h.getTo()+spec.states.size(), fa.createState());
				}
				fa.addTransition(st.getValue(arc_h.getFrom()), st.getValue(arc_h.getTo()), arc_h.getLabel()?"1":"0");
				fa.addTransition(st.getValue(arc_h.getFrom()+spec.states.size()), st.getValue(arc_h.getTo()), arc_h.getLabel()?"1":"0");

				Iterator<Pair<FAState,FAState>>brel_it=brel.iterator();
				while(brel_it.hasNext()){
					Pair<FAState,FAState> p=brel_it.next();
					if(p.getLeft().getowner()==spec&&p.getRight().getowner()==spec){
						if(!st.containsKey(p.getLeft().getID()))
							st.put(p.getLeft().getID(), fa.createState());
						if(!st.containsKey(p.getRight().getID()+spec.states.size()))
							st.put(p.getRight().getID()+spec.states.size(), fa.createState());
						fa.addTransition(st.getValue(p.getLeft().getID()), st.getValue(p.getRight().getID()+spec.states.size()), "0");
					}
				}
			}
			SCC s=new SCC(fa);
			TreeSet<FAState> Pre=new TreeSet<FAState>();
			TreeSet<FAState> Next=s.getResult();
			Pre.addAll(Next);
			while(!Next.isEmpty()){
				FAState state=Next.iterator().next();
				Next.remove(state);
				TreeSet<FAState> toAdd=new TreeSet<FAState>();
				toAdd.addAll(state.getPre());
				toAdd.removeAll(Pre);
				Pre.addAll(toAdd);
				Next.addAll(toAdd);
			}
			Next=null;
			Iterator<FAState> Pre_it=Pre.iterator();
			TreeSet<Integer> T=new TreeSet<Integer>();
			while(Pre_it.hasNext()){
				FAState state=Pre_it.next();
				T.add(st.getKey(state)%spec.states.size());
			}
			Tail.put(h.toString(), T);
		}
		TreeSet<Integer> intersection = new TreeSet<Integer>();
		intersection.addAll(Head.get(g.toString()));
		intersection.retainAll(Tail.get(h.toString()));
		
		if(debug){
			if(intersection.isEmpty()){
				debug("g_graph:"+g+", Head: "+Head.get(g.toString()));
				debug("h_graph:"+h+", Tail: "+Tail.get(h.toString()));
			}
		}
		
		return !intersection.isEmpty();
	}*/
	
	private Pair<Arc, TreeSet<Arc>> min(Pair<Arc, TreeSet<Arc>> supergraph){
		debug("before min:"+supergraph);
		TreeSet<Arc> result=new TreeSet<Arc>();
		Iterator<Arc> arc_it1 =supergraph.getRight().iterator();
		while(arc_it1.hasNext()){
			Arc cur=arc_it1.next();
			boolean canAdd=true;
			Iterator<Arc> arc_it2 =supergraph.getRight().iterator();
			while(arc_it2.hasNext()){
				Arc other=arc_it2.next();
				if(cur.getFrom()==other.getFrom()){
					if(!cur.getLabel()||other.getLabel()){
						if(cur.getTo()!=other.getTo()){
							if(frel.contains(new Pair<FAState,FAState>(new FAState(cur.getTo(),spec),new FAState(other.getTo(), spec )))){
								if(!frel.contains(new Pair<FAState,FAState>(new FAState(other.getTo(),spec),new FAState(cur.getTo(), spec )))){
									canAdd=false;
									break;
								}else if(cur.getTo()<other.getTo()){
									canAdd=false;
									break;
								}
							}
						}
					}
				}
			}			
			if(canAdd){
				result.add(cur);
			}
		}
		debug("after min:"+new Pair<Arc, TreeSet<Arc>>(supergraph.getLeft(),result));
		return new Pair<Arc, TreeSet<Arc>>(supergraph.getLeft(),result);
	}

	private boolean contains(Set<Pair<FAState,FAState>> rel, FAState s1, FAState s2){
		return rel.contains(new Pair<FAState,FAState>(s1,s2));
	}
	private Pair<Arc, TreeSet<Arc>> min_plus(Pair<Arc, TreeSet<Arc>> supergraph){
		debug("before min:"+supergraph);
		TreeSet<Arc> result=new TreeSet<Arc>();
		Iterator<Arc> arc_it1 =supergraph.getRight().iterator();
		while(arc_it1.hasNext()){
			Arc cur=arc_it1.next();
			boolean canAdd=true;
			Iterator<Arc> arc_it2 =supergraph.getRight().iterator();
			while(arc_it2.hasNext()){
				Arc other=arc_it2.next();
				if(cur.getFrom()==other.getFrom() && cur.getTo()==other.getTo())
					continue;
				if(contains(brel, new FAState(cur.getFrom(),spec),new FAState(other.getFrom(), spec )) &&
				   (other.getFrom()>=cur.getFrom()||!contains(brel, new FAState(other.getFrom(),spec),new FAState(cur.getFrom(), spec )))){
					if(!cur.getLabel()||other.getLabel()){
						if(contains(frel, new FAState(cur.getTo(),spec),new FAState(other.getTo(), spec )) &&
						   (other.getTo()>=cur.getTo()||!contains(frel, new FAState(other.getTo(),spec),new FAState(cur.getTo(), spec )))){
									canAdd=false;
									break;
						}
					}
				}
			}
			if(canAdd){
				result.add(cur);
			}
		}
		debug("after min:"+new Pair<Arc, TreeSet<Arc>>(supergraph.getLeft(),result));
		return new Pair<Arc, TreeSet<Arc>>(supergraph.getLeft(),result);
	}
	
	
	private ArrayList<Pair<Arc, TreeSet<Arc>>> buildSingleCharacterSuperGraphs(){
		ArrayList<Pair<Arc, TreeSet<Arc>>> supergraphs=new ArrayList<Pair<Arc, TreeSet<Arc>>>();
		
		Iterator<String> symbol_it=system.getAllTransitionSymbols().iterator();
		while(symbol_it.hasNext()){
			TreeSet<Arc> graph=new TreeSet<Arc>();
			
			String sym=symbol_it.next();
			Iterator<FAState> from_it=spec.states.iterator();
			while(from_it.hasNext()){
				FAState from=from_it.next();
				if(from.getNext(sym)!=null){
					Iterator<FAState> to_it=from.getNext(sym).iterator();
					while(to_it.hasNext()){
						FAState to=to_it.next();
						if(spec.F.contains(from)||spec.F.contains(to)){
							graph.add(new Arc(from.id,true,to.id));
						}else{
							graph.add(new Arc(from.id,false,to.id));
						}
					}
				}
			}

			from_it=system.states.iterator();
			while(from_it.hasNext()){
				FAState from=from_it.next();
				if(from.getNext(sym)!=null){
					Iterator<FAState> to_it=from.getNext(sym).iterator();
					while(to_it.hasNext()){
						FAState to=to_it.next();
						Arc left_arc;
						if(system.F.contains(from)||system.F.contains(to)){
							left_arc=new Arc(from.id,true,to.id);
						}else{
							left_arc=new Arc(from.id,false,to.id);
						}
						Pair<Arc, TreeSet<Arc>> supergraph=new Pair<Arc, TreeSet<Arc>>(left_arc,graph);
						ArrayList<Pair<Arc, TreeSet<Arc>>> toRemove=new ArrayList<Pair<Arc, TreeSet<Arc>>>();
						boolean canAdd=true;
						Iterator<Pair<Arc, TreeSet<Arc>>> old_it=supergraphs.iterator();
						while(old_it.hasNext()){
							Pair<Arc, TreeSet<Arc>> old=old_it.next();
							if(smallerThan(old, supergraph)){
								canAdd=false;
								break;
							}else if(smallerThan(supergraph, old)){
								if(debug)
									stringmap.remove(old);
								toRemove.add(old);
							}
						}
						if(canAdd){
							if(opt2)
								if(!backward)
									supergraph=min(supergraph);
								else
									supergraph=min_plus(supergraph);

							supergraphs.add(supergraph);
							supergraphs.removeAll(toRemove);
							if(debug)
								stringmap.put(supergraph, sym);
						}
					}
				}
			}
		}
		return supergraphs;
	}

	private Pair<Arc, TreeSet<Arc>> compose(Pair<Arc, TreeSet<Arc>> g, Pair<Arc, TreeSet<Arc>> h){
		TreeSet<Arc> f=new TreeSet<Arc>();
		Iterator<Arc> arc_g_it=g.getRight().iterator();
		while(arc_g_it.hasNext()){
			Arc arc_g=arc_g_it.next();
			Iterator<Arc> arc_h_it=h.getRight().iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				boolean connected=false;
				if(backward &&contains(brel, new FAState(arc_g.getTo(),spec),new FAState(arc_h.getFrom(), spec )))
					connected=true;
				else if(!backward&&arc_g.getTo()==arc_h.getFrom())
					connected=true;
				if(connected){
					if(arc_g.getLabel()||arc_h.getLabel()){
						f.add(new Arc(arc_g.getFrom(),true,arc_h.getTo()));
						f.remove(new Arc(arc_g.getFrom(),false,arc_h.getTo()));
					}else{
						if(!f.contains(new Arc(arc_g.getFrom(),true,arc_h.getTo()))){
							f.add(new Arc(arc_g.getFrom(),false,arc_h.getTo()));
						}
					}
				}				
			}			
		}
		if(g.getLeft().getLabel()||h.getLeft().getLabel())
			return new Pair<Arc, TreeSet<Arc>>(new Arc(g.getLeft().getFrom(),true,h.getLeft().getTo()),f);
		else
			return new Pair<Arc, TreeSet<Arc>>(new Arc(g.getLeft().getFrom(),false,h.getLeft().getTo()),f);
	}

	
	boolean smallerThan(Pair<Arc, TreeSet<Arc>> old, Pair<Arc, TreeSet<Arc>> old2){
		Arc old_arc=old.getLeft();
		Arc old2_arc=old2.getLeft();
		if(!(old_arc.getFrom()==old2_arc.getFrom() && (old_arc.getLabel() ||!old2_arc.getLabel()) &&
				frel.contains(new Pair<FAState,FAState>(new FAState(old2_arc.getTo(),system),new FAState(old_arc.getTo(),system))))){
			return false;
		}
		
		Iterator<Arc> arc_g_it=old.getRight().iterator();
		while(arc_g_it.hasNext()){
			Arc arc_g=arc_g_it.next();
			boolean has_larger=false;
			Iterator<Arc> arc_h_it=old2.getRight().iterator();
			while(arc_h_it.hasNext()){
				Arc arc_h=arc_h_it.next();
				boolean fromOkay=false;
				if(backward &&brel.contains(new Pair<FAState,FAState>(new FAState(arc_g.getFrom(),spec),new FAState(arc_h.getFrom(), spec ))))
					fromOkay=true;
				else if(!backward&&arc_g.getFrom()==arc_h.getFrom())
					fromOkay=true;
				if(fromOkay){
					if(!arc_g.getLabel()||arc_h.getLabel()){
						if(frel.contains(new Pair<FAState,FAState>(new FAState(arc_g.getTo(),spec),new FAState(arc_h.getTo(),spec)))){
							has_larger=true;
							break;
						}
					}
				}
			}			
			if(!has_larger){
				return false;
			}
		}
		return true;
	}
	public void inclusionTest(){
		this.computeSim();

		debug("Aut A:");
		debug(system.toString());
		debug("Aut B:");
		debug(spec.toString());
		debug("FSim="+showSim(frel));
		if(backward)
		debug("BSim="+showSim(brel));
		if(!C1&&!C2)
			this.included=included();
		else
			this.included=included_plus();
	}
	public long getCpuTime( ) {
	    ThreadMXBean bean = ManagementFactory.getThreadMXBean( );
	    return bean.isCurrentThreadCpuTimeSupported( ) ?
	        bean.getCurrentThreadCpuTime( ) : 0L;
	}
	
	public boolean isIncluded(){
		return included;
	}
	@Override
	public void run(){

		runTime=getCpuTime();
			inclusionTest();
		runTime=getCpuTime()-runTime;
	}
	
	public long getRunTime(){
		return runTime;
	}


	void computeSim(){

		ArrayList<FAState> all_states=new ArrayList<FAState>();
		HashSet<String> alphabet=new HashSet<String>();
		
		all_states.addAll(spec.states);
		alphabet.addAll(spec.alphabet);

		all_states.addAll(system.states);
		alphabet.addAll(system.alphabet);
		
		
		FAState[] states=all_states.toArray(new FAState[0]);		
		boolean[] isFinal = new boolean[states.length];
		boolean[] isInit = new boolean[states.length];
		boolean[][] fsim = new boolean[states.length][states.length];
		boolean[][] bsim = new boolean[states.length][states.length];
		// sim[u][v]=true iff v in sim(u) iff v simulates u
		for(int i=0;i<states.length;i++){			
			isFinal[i] = states[i].getowner().F.contains(states[i]);
			isInit[i] =states[i].getowner().getInitialState().compareTo(states[i])==0;
		}
		for(int i=0;i<states.length;i++){
			for(int j=i;j<states.length;j++){
				fsim[i][j] = (!isFinal[i] || isFinal[j]) && states[j].fw_covers(states[i]);
				fsim[j][i] = (isFinal[i] || !isFinal[j]) && states[i].fw_covers(states[j]);
				if(backward){
					bsim[i][j] = (!isInit[i] || isInit[j]) && (!isFinal[i] || isFinal[j]) && states[j].bw_covers(states[i]);
					bsim[j][i] = (isInit[i] || !isInit[j]) && (isFinal[i] || !isFinal[j]) && states[i].bw_covers(states[j]);
				}
			}
		}
		Simulation sim = new Simulation();
		frel = sim.FastFSimRelNBW(spec,system,fsim);
		if(quotient){
			spec=quotient(spec, frel);
			system=quotient(system, frel);
		}

		if(backward){
			all_states.clear();
			all_states.addAll(spec.states);
			all_states.addAll(system.states);
			states=all_states.toArray(new FAState[0]);		
			isFinal = new boolean[states.length];
			isInit = new boolean[states.length];
			bsim = new boolean[states.length][states.length];
			for(int i=0;i<states.length;i++){			
				isFinal[i] = states[i].getowner().F.contains(states[i]);
				isInit[i] =states[i].getowner().getInitialState().compareTo(states[i])==0;
			}
			for(int i=0;i<states.length;i++){
				for(int j=i;j<states.length;j++){
					bsim[i][j] = (!isInit[i] || isInit[j]) && (!isFinal[i] || isFinal[j]) && states[j].bw_covers(states[i]);
					bsim[j][i] = (isInit[i] || !isInit[j]) && (isFinal[i] || !isFinal[j]) && states[i].bw_covers(states[j]);
				}
			}
			brel = sim.FastBSimRelNBW(spec,system,bsim);
		}

	}	
	
	TreeMap<Pair<Arc, TreeSet<Arc>>, String> stringmap=new TreeMap<Pair<Arc, TreeSet<Arc>>, String>(new SuperGraphComparator());

	private boolean included(){
		
		ArrayList<Pair<Arc, TreeSet<Arc>>> Q1=this.buildSingleCharacterSuperGraphs();
		debug("Init:"+Q1);
		Iterator<Pair<Arc, TreeSet<Arc>>> Q1_it=Q1.iterator();
		while(Q1_it.hasNext()){
			Pair<Arc, TreeSet<Arc>> g=Q1_it.next();
			if(!double_graph_test(g, g, spec.getInitialState().id)){
				return false;
			}
		}
		TreeSet<Pair<Arc, TreeSet<Arc>>> Next=new TreeSet<Pair<Arc, TreeSet<Arc>>>(new SuperGraphComparator());
		TreeSet<Pair<Arc, TreeSet<Arc>>> Processed=new TreeSet<Pair<Arc, TreeSet<Arc>>>(new SuperGraphComparator());
		Next.addAll(Q1);
		while(!Next.isEmpty()){
			if(stop)
				break;
			debug("");
			debug("Begin while loop");
			debug("Processed:"+Processed);
			debug("Next:"+Next);
			
			Pair<Arc, TreeSet<Arc>> g=Next.first();
			debug("Pick "+g+"from Next");
						
			Iterator<Pair<Arc, TreeSet<Arc>>> Processed_it=Processed.iterator();
			while(Processed_it.hasNext()){
				Pair<Arc, TreeSet<Arc>> h=Processed_it.next();
				if(!double_graph_test(g, h, spec.getInitialState().id)||!double_graph_test(h, g, spec.getInitialState().id))
					return false;
			}
			Next.remove(g);
			Processed.add(g);
			debug("Move"+g+"from Next to Processed");
			
			Q1_it=Q1.iterator();
			while(Q1_it.hasNext()){
				ArrayList<Pair<Arc, TreeSet<Arc>>> toRemove=new ArrayList<Pair<Arc, TreeSet<Arc>>>();
				Pair<Arc, TreeSet<Arc>> h=Q1_it.next();
				if(composable(g,h)){
				Pair<Arc, TreeSet<Arc>> f=compose(g, h);
				boolean discard=false;

				debug("Produce f:"+f +"=Min_s("+g+"("+stringmap.get(g)+");"+h+"("+stringmap.get(h)+"))");
				if(debug)
					stringmap.put(f, stringmap.get(g)+stringmap.get(h));
				Processed_it=Processed.iterator();
				while(Processed_it.hasNext()){
					Pair<Arc, TreeSet<Arc>> p=Processed_it.next();
					if(smallerThan(f, p)){
						toRemove.add(p);
					}
					if(smallerThan(p, f)){
						debug("Discard "+f+"["+stringmap.get(f)+"] because it is larger than "+p+"["+stringmap.get(p)+"]");

						discard=true;
						break;
					}
				}
				if(discard)
					continue;
				Iterator<Pair<Arc, TreeSet<Arc>>> Next_it=Next.iterator();
				while(Next_it.hasNext()){
					Pair<Arc, TreeSet<Arc>> p=Next_it.next();
					if(smallerThan(f, p)){
						toRemove.add(p);
					}
					if(smallerThan(p, f)){
						debug("Discard "+f+"["+stringmap.get(f)+"] because it is larger than "+p+"["+stringmap.get(p)+"]");
						discard=true;
						break;
					}
				}
				if(discard)
					continue;
				if(!double_graph_test(f, f, spec.getInitialState().id))
					return false;
				Processed.removeAll(toRemove);
				Next.removeAll(toRemove);
				if(opt2){
					String indexStr=stringmap.get(f);
					stringmap.remove(f);
					if(!backward)
						f=min(f);
					else
						f=min_plus(f);
					stringmap.put(f,indexStr);
				}
				Next.add(f);
				debug("Add "+f+"("+stringmap.get(f)+") to Next");
				
				}
			}
		}			
		return true;
	}

	
	private boolean included_plus(){
		ArrayList<Pair<Arc, TreeSet<Arc>>> Q1=this.buildSingleCharacterSuperGraphs();
		Iterator<Pair<Arc, TreeSet<Arc>>> Q1_it=Q1.iterator();
		ArrayList<Pair<Arc, TreeSet<Arc>>> Q1_toRemove=new ArrayList<Pair<Arc, TreeSet<Arc>>>();
		debug("Init:");
		while(Q1_it.hasNext()){
			Pair<Arc, TreeSet<Arc>> g=Q1_it.next();
			boolean c1=C1(g);
			boolean c2=C2(g);
			
			String label="";
			if(!c2){
				R.add(g);
				label+="R";
			}
			if(!c1){
				L.add(g);
				label+="L";
			}
			if(c1&&c2){
				Q1_toRemove.add(g);
			}else
				debug(label+" "+g);
		}
		TreeSet<Pair<Arc, TreeSet<Arc>>> Next=new TreeSet<Pair<Arc, TreeSet<Arc>>>(new SuperGraphComparator());
		TreeSet<Pair<Arc, TreeSet<Arc>>> Processed=new TreeSet<Pair<Arc, TreeSet<Arc>>>(new SuperGraphComparator());
		Next.addAll(Q1);
		Next.removeAll(Q1_toRemove);

		while(!Next.isEmpty()&&L.size()!=0&&R.size()!=0){
			if(stop)
				break;
			Pair<Arc, TreeSet<Arc>> g=Next.first();
			debug("");
			debug("Begin while loop");
			debug("Processed:"+Processed);
			debug("Next:"+Next);
			Next.remove(g);
			if(L.contains(g)&&!R.contains(g)){
				debug("Pick and remove"+g+"with label L from Next");
				Iterator<Pair<Arc, TreeSet<Arc>>> Processed_it=Processed.iterator();
				while(Processed_it.hasNext()){
					Pair<Arc, TreeSet<Arc>> h=Processed_it.next();
					if(R.contains(h))
					if(!double_graph_test(g, h, spec.getInitialState().id))
						return false;
				}
				Processed.add(g);
				L.remove(g);
				debug("Add "+g+" to Processed with label N (no label)");

				Q1_it=Q1.iterator();
				while(Q1_it.hasNext()){
					ArrayList<Pair<Arc, TreeSet<Arc>>> toRemove=new ArrayList<Pair<Arc, TreeSet<Arc>>>();
					Pair<Arc, TreeSet<Arc>> h=Q1_it.next();
					if(composable(g,h)){
						Pair<Arc, TreeSet<Arc>> f=compose(g, h);
						if(C1(f))
							continue;
						if(opt2)
							if(!backward)
								f=min(f);
							else
								f=min_plus(f);
						
						boolean discard=false;
						debug("Produce f:"+f +"=Min_s("+g+";"+h+")");
						if(debug&&!stringmap.containsKey(f))
							stringmap.put(f, stringmap.get(g)+stringmap.get(h));
						Processed_it=Processed.iterator();
						while(Processed_it.hasNext()){
							Pair<Arc, TreeSet<Arc>> k=Processed_it.next();
							if(smallerThan(f, k)){
								toRemove.add(k);
							}
							if(smallerThan(k, f)){
								discard=true;
								if(!L.contains(k)){
									L.add(k);
									debug("Upgrade "+k +" with label L");
									Iterator<Pair<Arc, TreeSet<Arc>>> Processed_it2=Processed.iterator();
									while(Processed_it2.hasNext()){
										Pair<Arc, TreeSet<Arc>> h2=Processed_it2.next();
										if(R.contains(h2))
										if(!double_graph_test(k, h2, spec.getInitialState().id))
											return false;
									}
								}
								break;
							}
						}
						if(discard)
							continue;
						Iterator<Pair<Arc, TreeSet<Arc>>> Next_it=Next.iterator();
						while(Next_it.hasNext()){
							Pair<Arc, TreeSet<Arc>> k=Next_it.next();
							if(smallerThan(f, k)){
								toRemove.add(k);
							}
							if(smallerThan(k, f)){
								discard=true;
								break;
							}
						}
						if(discard)
							continue;
						Processed.removeAll(toRemove);
						Next.removeAll(toRemove);
						L.removeAll(toRemove);
						R.removeAll(toRemove);
						L.add(f);
						Next.add(f);
						debug("Label "+f+" with L and add it to Next");
					}
				}				
				
			}else if(!L.contains(g)&&R.contains(g)){
				debug("Pick and remove"+g+"with label R from Next");
				Iterator<Pair<Arc, TreeSet<Arc>>> Processed_it=Processed.iterator();
				while(Processed_it.hasNext()){
					Pair<Arc, TreeSet<Arc>> h=Processed_it.next();
					if(L.contains(h))
					if(!double_graph_test(h, g, spec.getInitialState().id))
						return false;
				}
				Processed.add(g);
				debug("Label "+g+" with R and add it to Processed");
				Q1_it=Q1.iterator();
				while(Q1_it.hasNext()){
					ArrayList<Pair<Arc, TreeSet<Arc>>> toRemove=new ArrayList<Pair<Arc, TreeSet<Arc>>>();
					Pair<Arc, TreeSet<Arc>> h=Q1_it.next();
					if(composable(g,h)){
						Pair<Arc, TreeSet<Arc>> f=compose(g, h);
						if(C2(f))
							continue;
						if(opt2)
							if(!backward)
								f=min(f);
							else
								f=min_plus(f);
						
						boolean discard=false;
						debug("Produce f:"+f +"=Min_s("+g+";"+h+")");
						if(debug&&!stringmap.containsKey(f))
							stringmap.put(f, stringmap.get(g)+stringmap.get(h));
						Processed_it=Processed.iterator();
						while(Processed_it.hasNext()){
							Pair<Arc, TreeSet<Arc>> p=Processed_it.next();
							if(smallerThan(f, p)){
								toRemove.add(p);
							}
							if(smallerThan(p, f)){
								discard=true;
								break;
							}
						}
						if(discard)
							continue;
						Iterator<Pair<Arc, TreeSet<Arc>>> Next_it=Next.iterator();
						while(Next_it.hasNext()){
							Pair<Arc, TreeSet<Arc>> p=Next_it.next();
							if(smallerThan(f, p)){
								toRemove.add(p);
							}
							if(smallerThan(p, f)){
								discard=true;
								break;
							}
						}
						if(discard)
							continue;
						Processed.removeAll(toRemove);
						Next.removeAll(toRemove);
						L.removeAll(toRemove);
						R.removeAll(toRemove);
						debug("Label "+f+" with R and add it to Next");
						R.add(f);
						Next.add(f);
					}
				}				
			}else if(L.contains(g)&&R.contains(g)){
				debug("Pick and remove"+g+"with label LR from Next");
				if(!double_graph_test(g, g, spec.getInitialState().id))
					return false;

				Iterator<Pair<Arc, TreeSet<Arc>>> Processed_it=Processed.iterator();
				while(Processed_it.hasNext()){
					Pair<Arc, TreeSet<Arc>> h=Processed_it.next();
					if(L.contains(h))
						if(!double_graph_test(h, g, spec.getInitialState().id))
							return false;
					if(R.contains(h))
						if(!double_graph_test(g, h, spec.getInitialState().id))
							return false;
					
				}
				Processed.add(g);
				L.remove(g);
				debug("Label "+g+" with R and add it to Next");
				Q1_it=Q1.iterator();
				while(Q1_it.hasNext()){
					ArrayList<Pair<Arc, TreeSet<Arc>>> toRemove=new ArrayList<Pair<Arc, TreeSet<Arc>>>();
					Pair<Arc, TreeSet<Arc>> h=Q1_it.next();
					if(composable(g,h)){
						Pair<Arc, TreeSet<Arc>> f=compose(g, h);
						boolean c1=C1(f);
						boolean c2=C2(f);
						if(c1&&c2)
							continue;
						if(opt2)
							if(!backward)
								f=min(f);
							else
								f=min_plus(f);
						
						boolean discard=false;
						debug("Produce f:"+f +"=Min_s("+g+";"+h+")");
						if(debug&&!stringmap.containsKey(f))
							stringmap.put(f, stringmap.get(g)+stringmap.get(h));

						Processed_it=Processed.iterator();
						while(Processed_it.hasNext()){
							Pair<Arc, TreeSet<Arc>> k=Processed_it.next();
							if(smallerThan(f, k)){
								toRemove.add(k);
							}
							if(smallerThan(k, f)){
								if(!c1){
									if(!L.contains(k)){
										L.add(k);
										Iterator<Pair<Arc, TreeSet<Arc>>> Processed_it2=Processed.iterator();
										while(Processed_it2.hasNext()){
											Pair<Arc, TreeSet<Arc>> h2=Processed_it2.next();
											if(R.contains(h2))
											if(!double_graph_test(k, h2, spec.getInitialState().id))
												return false;
										}
									}
								}
								discard=true;
								break;
							}
						}
						if(discard)
							continue;
						Iterator<Pair<Arc, TreeSet<Arc>>> Next_it=Next.iterator();
						while(Next_it.hasNext()){
							Pair<Arc, TreeSet<Arc>> k=Next_it.next();
							if(smallerThan(f, k)){
								toRemove.add(k);
							}
							if(smallerThan(k, f)){
								discard=true;
								break;
							}
						}
						if(discard)
							continue;
						Processed.removeAll(toRemove);
						Next.removeAll(toRemove);
						L.removeAll(toRemove);
						R.removeAll(toRemove);
						if(c1){R.add(f);debug("Label "+f+" with R and add it to Next");}
						else if(c2){L.add(f);debug("Label "+f+" with L and add it to Next");}
						else {R.add(f);L.add(f);debug("Label "+f+" with LR and add it to Next");}
						Next.add(f);
					}
				}				
			}
		}			
		return true;
	}
		
	
	private boolean composable(Pair<Arc, TreeSet<Arc>> g,
			Pair<Arc, TreeSet<Arc>> h) {
//		if(backward && 
//		contains(brel, new FAState(g.getLeft().getTo(),system), new FAState(h.getLeft().getFrom(),system)))
//			return true;
//		else if(!backward&&g.getLeft().getTo()==h.getLeft().getFrom())
//			return true;
//		else
//			return false;
		if(g.getLeft().getTo()==h.getLeft().getFrom())
			return true;
		else
			return false;
		
	}
	public void stopIt(){
		stop=true;
	}
}
